/**** SimCore since 2002-05-15 by Kenji KISE ****/ 
/************************************************/
#include "define.h"
#include <unistd.h> 

/************************************************/
int system_manager::translate_fd(DATA_TYPE fd){
  for(int i=0; i<FD_MAPPING_MAX; i++){ 
    if(fd_table[i] == fd) return i;
  }
  printf("*** Error in translate_fd\n");
  exit(1);
  return 0;
}

/************************************************/
system_manager::system_manager(Env *e){
  ev = e;
  uniq    = 0;
  running = 1;
  
  /** for example, fd_table[6]=7        **/
  /** maps real fd(6) to logical fd(7). **/
  /***************************************/
  fd_table[0] = 0; // STDIN
  fd_table[1] = 1; // STDOUT
  fd_table[2] = 2; // STDERR
  for(int i=3; i<FD_MAPPING_MAX; i++)
    fd_table[i] = i + 2 - ev->sc->fd_adjust;
}

/************************************************/
void system_manager::execute_pal(instruction *p){
  INST_TYPE ir = p->ir;
  int_64t ret=0;
  int F = ev->sc->syscall_flag;
  data_t *r = ev->as->r;
  
  if(F) printf("%8lld: ", ev->e->retired_inst);
  
  if((ir & 0xFFFF)!=0x0083){
    switch((ir & 0xFFFF)){
    case 0x009e: // rdunique(Read Unique Value)
      r[0]=(uniq);
      if(F){
	printf("call_pal rduniq=%llx\n", uniq);
      }
      return;
    case 0x009f: // wrunique(Write Unique Value)
      uniq  = r[16];
      if(F){
	printf("call_pal wrruniq=%llx\n", 
	       r[16]);
      }
      return;
    default:
      if(F){
	printf("*** PAL(ir & 0xFFFF)!=0x83 F=%x",
	       (ir & 0xffff));
	printf("  TC:%lld default. %llx\n",
	       ev->e->retired_inst, r[0]);
      }
      return; 
    }
  }
  
  if(r[0]==0){
    r[19]=(0);
    return;
  }

  /**********************************************/
  /**********************************************/
  if(r[0]==ASYS_UNIMPLEMENT){
    ret = -1;
    r[0]=(0);
    r[19]=(ret);
    if(F){
      printf("SYS_226 = %lld\n", ret);
    }
  }

  /**********************************************/
  else if(r[0]==ASYS_read){
    char *buf = (char*)calloc(1, (size_t)r[18]);
    if(buf==NULL){
      printf("*** malloc error: ");
      printf("in syscall.cc::read\n");
    }
    
    int fd_t = translate_fd(r[16]);
    if(ev->sc->fd_stdin_flag && fd_t==0) 
      fd_t=ev->sc->fd_stdin; /* STDIN */
    
    ret = read(fd_t, buf, r[18]);
    
    // This is better. & 
    // simple scalar imple!
    /*      for(int i=0; i<ret; i++){  */
    for(int i=0; i<(int)r[18]; i++){ 
      DATA_TYPE d = buf[i];
      data_t adr; adr=(r[17]+i);
      data_t dat; dat=(d);
      ev->mem->st_nbyte(1, &adr, &dat);
    }
    r[0]=(ret);
    r[19]=(0);
    free(buf);
    
    if(F){
      printf("read");
      printf("(logic %lld(real%d),%llx,0x%llx)",
	     r[16], 
	     fd_t, r[17], r[18]);
      printf(" = %llx\n", ret);
    }
  }
  
  /**********************************************/
  else if(r[0]==ASYS_write){
    char *buf = (char *)malloc((size_t)r[18]);
    if(buf==NULL){
      printf("*** malloc error: ");
      printf("in syscall.cc::write\n");
    }
    
    for(int i=0; i<(int)r[18]; i++){
      data_t d;
      data_t adr;
      adr=(r[17]+i);
      ev->mem->ld_nbyte(1, &adr, &d);
      buf[i] = (char)d;
    }
    
    int fd_t = translate_fd(r[16]);
    
    ret = write(fd_t, (void *)buf,
		(long)r[18]);
    r[0]=(ret);
    r[19]=(0);
    free(buf);
    
    if(F){
      printf("write");
      printf("(%lld(%d),0x%llx,%lld)=%lld\n",
	     r[16],
	     fd_t, r[17], r[18], ret);
    }
  }
  
  /**********************************************/
  else if(r[0]==ASYS_unlink){
    char buf[512]; /** file name **/
    for(int i=0; i<=512; i++){
      data_t d;
      data_t adr;
      adr=(r[16]+i);
      ev->mem->ld_nbyte(1, &adr, &d);
      buf[i] = (char)d;
      if(buf[i]=='\0') break;
    }
    if(ev->sc->fd_adjust){ /** rename **/
      buf[1]='S';
      buf[2]='A';
      buf[3]='0'+ ev->sc->fd_adjust;
    }
    
    ret = unlink(buf);
    if(ret==-1){
      r[0]=(errno);
      r[19]=(ret);
    }
    else{
      r[19]=(0);
    }
    
    if(F){
      printf("unlink(%s) = %lld\n", buf, ret);
    }
  }

  /**********************************************/
  else if(r[0]==ASYS_open){
    int fd_t;
    char buf[512]; /** file name **/
    for(int i=0; i<=512; i++){
      data_t d;
      data_t adr;
      adr=(r[16]+i);
      ev->mem->ld_nbyte(1, &adr, &d);
      buf[i] = (char)d;
      if(buf[i]=='\0') break;
    }
    /** RDONLY or RDWR **/
    if(r[17]==0 || r[17]==2){
      fd_t = open(buf, r[17], r[18]);
    }
    else{
      if(ev->sc->fd_adjust){ /** rename **/
	buf[1]='S';
	buf[2]='A';
	buf[3]='0'+ ev->sc->fd_adjust;
      }
      int flag = (int)r[17] | O_CREAT;
      fd_t = open(buf, flag, r[18]);
    }
    
    /* mapping from the real to the logical */
    if(fd_t==-1){
      ret = -1;
      r[0]=(errno);
      r[19]=(~0ull);
    }
    else{
      ret = fd_table[fd_t];
      r[0]=(ret);
      r[19]=(0);
    }
    
    if(F){
      printf("open(\"%s\", 0x%llx, 0x%llx)",
	     buf, r[17], r[18]);
      printf("= logic %lld(real %d)\n",
	     ret, fd_t);
    }
  }

  /**********************************************/
  else if(r[0]==ASYS_sigreturn){
    uint_64t pt[OSF_SIGCONTEXT_SIZE/8];
    char *buf = (char *)pt;
    int i;

    data_t d;
    data_t adr;
    for(i=0; i<OSF_SIGCONTEXT_SIZE; i++){
      adr=(r[16] + i);
      ev->mem->ld_nbyte(1, &adr, &d);
      buf[i] = (char)d;
    }
    
    for(i=0; i<31; i++) r[i]=(pt[i+4]);
    for(i=0; i<31; i++){
      ev->as->f[i]=(pt[i+37]);
    }
    ev->as->pc=(pt[2]);  /** Update PC **/
    
    if(F){ 
      printf("sigreturn(0x%llx),Npc:0x%llx\n", 
	     r[16], ev->as->pc);
    }
  }

  /**********************************************/
  else if(r[0]==ASYS_close){
    int fd_t = translate_fd(r[16]);
    
    if(fd_t!=0 && fd_t!=1 && fd_t!=2){
      ret = close(fd_t);
      r[0]=(ret);
    }
    r[19]=(0);
    if(F){
      printf("close(%lld(%d)) = %lld\n", 
	     r[16], fd_t, ret);
    }
  }
  
  /**********************************************/
  else if(r[0]==ASYS_getpagesize){
    ret = 8192; /* getpagesize(); */
    r[0]=(ret);
    r[19]=(0);
    if(F){
      printf("getpagesize() = 0x%llx\n", ret);
    }
  }
  
  /**********************************************/
  else if(r[0]==ASYS_sigprocmask){
    ret = 0;
    r[0]=(ret);
    r[19]=(ret);
    if(F){
      printf("sigprocmask(%lld,%llx) = %lld\n",
	     r[16], r[17], ret);
    }
  }
  
  /**********************************************/
  else if(r[0]==ASYS_getrusage){ // Not Checked 
    struct rusage ru;
    //  ret=getrusage((enum __rusage_who)r[16],
    ret=getrusage((int)r[16],
		  &ru);
    
    for(int i=0; i<OSF_RUSAGE_SIZE; i++){
      data_t adr; adr=(r[17]+i);
      data_t dat; dat=(0);
      ev->mem->st_nbyte(1, &adr, &dat);
    }
    
    r[0]=(ret);
    r[19]=(0);
    
    if(F){
      printf("getrusage(%lld, %llx)=%lld\n", 
	     r[16], r[17], ret);
    }
  }
  
  /**********************************************/
  else if(r[0]==ASYS_gettimeofday){
    r[19]=(0);
    if(F){ 
      printf("gettimeofday(%lld,%llx) = %lld\n", 
	     r[16], r[17], ret);
    }
  }

  /**********************************************/
  else if(r[0]==ASYS_fcntl){
    ret = 0;
    r[0]=(ret);
    r[19]=(0);
    if(F){
      printf("fcntl(%lld,%llx) = %lld\n", 
	     r[16], r[17], ret);
    }
  }

  /**********************************************/
  else if(r[0]==ASYS_sigaction){
    ret = 0;
    r[0]=(ret);
    r[18]=(r[17]);
    r[19]=(0);
    if(F){
      printf("sigaction(%llx,%llx,%llx,%llx)", 
	     r[16], r[17],
	     r[18], r[19]);
      printf(" = %lld\n", ret);
    }
  }

  /**********************************************/
  else if(r[0]==ASYS_osf_getsysinfo){
    ret = 0;
    r[0]=(ret);
    r[19]=(0);
    if(F){ 
      printf("osf_getsysinfo");
      printf("(%llx,%llx,%llx,%llx)", 
	     r[16], r[17], r[18], r[19]);
      printf(" = %lld\n", ret);
    }
  }

  /**********************************************/
  else if(r[0]==ASYS_osf_setsysinfo){
    ret = 0;
    r[0]=(ret);
    //    r[20]= ret;
    r[19]=(0);
    if(F){
      printf("osf_setsysinfo");
      printf("(%llx,%llx,%llx,%llx)", 
	     r[16], r[17], r[18], r[19]);
      printf(" = %lld\n", ret);
    }
  }

  /**********************************************/
  else if(r[0]==ASYS_getuid){
    ret = getuid();
    r[0]=(ret);
    r[19]=(0);
    r[20]=(geteuid());
    if(F){
      printf("getuid() = %lld\n", ret);
    }
  }

  /**********************************************/
  else if(r[0]==ASYS_getgid){
    ret = getgid();
    r[0]=(ret);
    r[19]=(0);
    r[20]=(getegid());
    if(F){
      printf("getgid(0x%llx) = %lld\n", 
	     r[16], ret);
    }
  }

  /**********************************************/
  else if(r[0]==ASYS_brk){
    ret = r[16];
    r[0]=(ret);
    r[19]=(0);
    if(F){
      printf("brk(0x%llx) = %llx\n", 
	     r[16], ret);
    }
  }

  /**********************************************/
  else if(r[0]==ASYS_getpid){
    ret = getpid();
    r[0]=(ret);
    r[19]=(0);
    if(F){
      printf("getpid() = %llx\n", ret);
    }
  }
  
  /**********************************************/
  else if(r[0]==ASYS_fstat){
    int fd_t = translate_fd(r[16]);
    
    struct stat sbuf;
    ret = fstat(fd_t,  &sbuf);
    
    struct osf_statbuf asbuf;
    memset(&asbuf,0,sizeof(struct osf_statbuf));
    asbuf.a_st_dev     = sbuf.st_dev;
    asbuf.a_st_ino     = sbuf.st_ino;
    asbuf.a_st_mode    = sbuf.st_mode;
    asbuf.a_st_nlink   = sbuf.st_nlink;
    asbuf.a_st_uid     = sbuf.st_uid;
    asbuf.a_st_gid     = sbuf.st_gid;
    asbuf.a_st_rdev    = sbuf.st_rdev;
    asbuf.a_st_size    = sbuf.st_size;
    asbuf.a_st_atime   = sbuf.st_atime;
    asbuf.a_st_mtime   = sbuf.st_mtime;
    asbuf.a_st_ctime   = sbuf.st_ctime;
    asbuf.a_st_blksize = sbuf.st_blksize;
    asbuf.a_st_blocks  = sbuf.st_blocks;
    
    char *buf = (char *)&asbuf;
    
    for(int i=0; i<OSF_FSTAT_SIZE; i++){
      data_t adr; 
      adr=(r[17]+i);
      data_t dat; 
      dat=(buf[i]);
      ev->mem->st_nbyte(1, &adr, &dat);
    }
    
    r[0]=(ret);
    r[19]=(0);
    if(F){
      printf("fstat");
      printf("(logic %lld(real %d), 0x%llx)", 
	     r[16],fd_t, r[17]);
      printf(" = %lld\n", ret);
    }
  }

  /**********************************************/
  else if(r[0]==ASYS_ioctl){
    int fd_t = translate_fd(r[16]);
    
    switch(r[17]){
    case 0x40067408:      // OSF_TIOCGRETP 
      {
	struct osf_sgttyb {
	  uint_08t sg_ispeed; /* input speed  */
	  uint_08t sg_ospeed; /* output speed */
	  uint_08t sg_erase;  /* erase char   */
	  uint_08t sg_kill;   /* kill char    */
	  uint_16t sg_flags;  /* mode flags   */
	};
	
	struct termios lbuf;
	ret = tcgetattr(fd_t, &lbuf);
	
	/*** store 6 byte, value of 0 ***/
	for(int i=0; i<6; i++){
	  data_t adr; adr=(r[18]+i);
	  data_t dat; dat=(0);
	  ev->mem->st_nbyte(1, &adr, &dat);
	}
	
	if(ret!=-1){
	  r[0]=(ret);
	  r[19]=(0); // A3
	}
	else{
	  r[0]=(errno);
	  r[19]=(ret);
	}
	break;
      }
      
    default:
      ret = 0;
      r[19]=(ret);
      if(F) printf("*** ioctl default\n");
      break;
    }
    
    if(F){
      printf("ioctl");
      printf("(logic%lld(r%d),0x%llx,0x%llx)", 
	     r[16], fd_t,
	     r[17], r[18]); 
      printf(" = %llx\n", ret);
    }
  }

  /**********************************************/
  else if(r[0]==ASYS_lseek){
    int fd_t = translate_fd(r[16]);
    if(ev->sc->fd_stdin_flag && fd_t==0) 
      fd_t=ev->sc->fd_stdin; /* STDIN */
    
    ret = lseek(fd_t, r[17], r[18]);
    if(ret==-1){
      r[0]=(errno);
      r[19]=(~0);
    }
    else{
      r[0]=(ret);
      r[19]=(0);
    }
    if(F){
      printf("lseek");
      printf("(logic%lld(real%d),%lld,%lld)\n",
	     r[16], fd_t, 
	     r[17], r[18]);
      printf(" = %llx\n", ret);
    }
  }

  /**********************************************/
  else if(r[0]==ASYS_getrlimit){
    ret = 0;
    r[19]=(ret);
    if(F){
      printf("%8lld: getrlimit(%lld,%lld,%lld)",
	     ev->e->retired_inst, r[16], 
	     r[17], r[18]);
      printf(" = %llx\n", ret);
    }
  }

  /**********************************************/
  else if(r[0]==ASYS_setrlimit){
    ret = 0;
    r[19]=(ret);
    if(F){
      printf("setrlimit");
      printf("(%lld,%lld,%lld) = %lld\n",
	     r[16], r[17], r[18], ret);
    }
  }

  /**********************************************/
  else if(r[0]==ASYS_exit){
    ret = 0;
    running = 0;
    if(F){
      printf("%8lld: exit() = %lld\n", 
	     ev->e->retired_inst, ret);
    }
  }
  
  /**********************************************/
  else{
    printf("syscall not implemented!!: %lld\n",
	   r[0]);
    r[0]=(0);
    r[19]=((uint_64t)-1);
  }
  fflush(stdout);
}
/************************************************/
